@@ -1,16 +1,11 @@
-"""
-Interactive explorers mostly based on Bokeh.
-"""
-import pandas as pd
+"""Interactive explorers mostly based on Bokeh."""
 import numpy as np
-import bokeh
-import wrappy
 from wasabi import msg as logger
-from bokeh.models import CustomJS, ColumnDataSource
+from bokeh.models import CustomJS, ColumnDataSource, CDSView, IndexFilter
 from bokeh.layouts import column, row
 from abc import ABC, abstractmethod
-from copy import deepcopy
 from hover import module_config
+from hover.utils.misc import current_time
 from .local_config import bokeh_hover_tooltip
 
 
@@ -43,6 +38,12 @@ class BokehForLabeledText(ABC):
     }
 
     def __init__(self, **kwargs):
+        """
+        Operations shared by all child classes.
+        (1) settle the figure settings by using child class defaults + kwargs overrides
+        (2) create a blank figure under such settings
+        (3) create widgets that child classes can override
+        """
         self.figure_settings = self.__class__.DEFAULT_FIGURE_KWARGS.copy()
         self.figure_settings.update(kwargs)
         self.reset_figure()
@@ -183,6 +184,12 @@ class BokehCorpusExplorer(BokehForLabeledText):
     }
 
     def __init__(self, df_raw, **kwargs):
+        """
+        Requires the input dataframe to contain:
+
+        (1) "x" and "y" columns for coordinates;
+        (2) a "text" column for data point tooltips.
+        """
         super().__init__(**kwargs)
 
         # prepare plot-ready dataframe for train set
@@ -195,8 +202,6 @@ class BokehCorpusExplorer(BokehForLabeledText):
         self.source = ColumnDataSource(self.df_raw)
         self._activate_search_on_corpus()
 
-        self.figure.circle("x", "y", source=self.source, **self.background_kwargs)
-
     def _activate_search_on_corpus(self):
         """
         Assuming that there will not be a labeled dev set.
@@ -217,9 +222,11 @@ class BokehCorpusExplorer(BokehForLabeledText):
 
     def plot(self, *args, **kwargs):
         """
-        Does nothing.
+        (Re)-plot the corpus.
+        Called just once per instance most of the time.
         """
-        pass
+        self.reset_figure()
+        self.figure.circle("x", "y", source=self.source, **self.background_kwargs)
 
 
 class BokehCorpusAnnotator(BokehForLabeledText):
@@ -229,7 +236,15 @@ class BokehCorpusAnnotator(BokehForLabeledText):
     """
 
     def __init__(self, df_working, **kwargs):
+        """
+        Requires the input dataframe to contain:
 
+        (1) "x" and "y" columns for coordinates;
+        (2) a "text" column for data point tooltips.
+
+        Optional:
+        (3) a "label" column that defaults to ABSTAIN.
+        """
         super().__init__(**kwargs)
 
         for _key in ["text", "x", "y"]:
@@ -300,17 +315,17 @@ class BokehCorpusAnnotator(BokehForLabeledText):
 
             self.reset_source()
             self.plot()
-            logger.good(f"Updated annotator plot")
+            logger.good(f"Updated annotator plot at {current_time()}")
 
         def export():
             # a callback on clicking self.annotator_export
-            import dill
+            from dill import dump
             from datetime import datetime
 
             timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
             filename = f"bokeh-annotated-df-{timestamp}.pkl"
             with open(filename, "wb") as f:
-                dill.dump(self.df_working, f)
+                dump(self.df_working, f)
             logger.good(f"Saved DataFrame to {filename}")
 
         self.annotator_apply.on_event(ButtonClick, apply)
@@ -350,6 +365,12 @@ class BokehMarginExplorer(BokehCorpusExplorer):
     }
 
     def __init__(self, df_raw, label_col_a, label_col_b, **kwargs):
+        """
+        On top of the requirements of the parent class,
+        the input dataframe should contain:
+
+        (1) label_col_a and label_col_b for "label margins".
+        """
         super().__init__(df_raw, **kwargs)
 
         for _key in ["text", label_col_a, label_col_b, "x", "y"]:
@@ -369,7 +390,6 @@ class BokehMarginExplorer(BokehCorpusExplorer):
         """
         Plot the margins about a single label.
         """
-        from bokeh.models import CDSView, IndexFilter
 
         # prepare plot settings
         axes = ("x", "y")
@@ -410,11 +430,20 @@ class BokehSnorkelExplorer(BokehCorpusExplorer):
     Plot text data points along with labeling function outputs.
     """
 
+    BACKGROUND_GLYPH_KWARGS = {"line_alpha": 0.6, "fill_alpha": 0.0, "size": 7}
+
     def __init__(self, df_raw, df_labeled, **kwargs):
+        """
+        On top of the requirements of the parent class,
+        the df_labeled input dataframe should contain:
+
+        (1) a "label" column for "ground truths".
+        """
         super().__init__(df_raw, **kwargs)
 
         # add 'label' column to df_raw
-        self.df_raw["label"] = "unlabeled"
+        if not "label" in self.df_raw.columns:
+            self.df_raw["label"] = module_config.ABSTAIN_DECODED
 
         # prepare plot-ready dataframe for dev set
         for _key in ["text", "label", "x", "y"]:
@@ -424,21 +453,23 @@ class BokehSnorkelExplorer(BokehCorpusExplorer):
         # initialize a list of LFs to enable toggles
         self.lfs = []
 
-    def _activate_search_on_corpus(self, source, background_kwargs):
+    def _activate_search_on_corpus(self):
         """
         Overriding the parent method, because there will be a labeled dev set.
         """
-        self.activate_search(
-            source, background_kwargs, altered_param=("size", 10, 5, 7)
+        self.background_kwargs = self.activate_search(
+            self.source, self.background_kwargs, altered_param=("size", 10, 5, 7)
         )
-        self.activate_search(
-            source, background_kwargs, altered_param=("fill_alpha", 0.6, 0.0, 0.3)
+        self.background_kwargs = self.activate_search(
+            self.source,
+            self.background_kwargs,
+            altered_param=("fill_alpha", 0.6, 0.0, 0.3),
         )
 
-    # def plot(self, lf, L_train, L_dev, include, **kwargs):
-    def plot(self, lf, L_raw=None, L_labeled=None, include=["C", "I", "M"], **kwargs):
+    def plot(self, lf, L_raw=None, L_labeled=None, include=("C", "I", "M"), **kwargs):
         """
         Plot a single labeling function.
+        :param lf: labeling function decorated by @hover.utils.snorkel_helper.labeling_function()
         """
         # keep track of added LF
         self.lfs.append(lf)
@@ -452,75 +483,75 @@ class BokehSnorkelExplorer(BokehCorpusExplorer):
         # prepare plot settings
         axes = ("x", "y")
         decoded_targets = [lf.label_decoder[_target] for _target in lf.targets]
-        legend = f"{', '.join(decoded_targets)} | {lf.name}"
-        template_kwargs = {"line_alpha": 0.6, "fill_alpha": 0.0, "size": 7}
-        template_kwargs.update(kwargs)
+        eff_kwargs = self.background_kwargs.copy()
+        eff_kwargs.update(kwargs)
+        eff_kwargs["legend_label"] = f"{', '.join(decoded_targets)} | {lf.name}"
 
         # create correct/incorrect/missed/hit subsets
         to_plot = []
         if "C" in include:
             to_plot.append(
-                {
-                    "source": self._source_correct(L_labeled),
-                    "marker": self.figure.square,
-                }
+                {"view": self._view_correct(L_labeled), "marker": self.figure.square}
             )
         if "I" in include:
             to_plot.append(
-                {"source": self._source_incorrect(L_labeled), "marker": self.figure.x}
+                {"view": self._view_incorrect(L_labeled), "marker": self.figure.x}
             )
         if "M" in include:
             to_plot.append(
                 {
-                    "source": self._source_missed(L_labeled, lf.targets),
+                    "view": self._view_missed(L_labeled, lf.targets),
                     "marker": self.figure.cross,
                 }
             )
         if "H" in include:
             to_plot.append(
-                {"source": self._source_hit(L_raw), "marker": self.figure.circle}
+                {"view": self._view_hit(L_raw), "marker": self.figure.circle}
             )
 
         # plot created subsets
         for _dict in to_plot:
-            _source = _dict["source"]
+            _view = _dict["view"]
             _marker = _dict["marker"]
-            _kwargs = deepcopy(template_kwargs)
-            self.activate_search(_source, _kwargs)
-            _marker(*axes, source=_source, legend_label=legend, **_kwargs)
+            _marker(*axes, source=self.source, view=_view, **eff_kwargs)
 
-    @wrappy.todo("Review whether it's appropriate to create a ColumnDataSource")
-    def _source_correct(self, L_labeled):
+    def _view_correct(self, L_labeled):
         """
         Determine the subset correctly labeled by a labeling function.
         """
-        indices = self.df_labeled["label"].values == L_labeled
-        return ColumnDataSource(self.df_labeled.iloc[indices])
 
-    @wrappy.todo("Review whether it's appropriate to create a ColumnDataSource")
-    def _source_incorrect(self, L_labeled):
+        indices_raw = self.df_labeled["label"].values == L_labeled
+        indices = np.where(indices_raw)[0].tolist()
+        view = CDSView(source=self.source, filters=[IndexFilter(indices)])
+        return view
+
+    def _view_incorrect(self, L_labeled):
         """
         Determine the subset incorrectly labeled by a labeling function.
         """
         disagreed = self.df_labeled["label"].values != L_labeled
-        attempted = L_labeled != module_config.ABSTAIN_ENCODED
-        indices = np.multiply(disagreed, attempted)
-        return ColumnDataSource(self.df_labeled.iloc[indices])
+        attempted = L_labeled != module_config.ABSTAIN_DECODED
+        indices_raw = np.multiply(disagreed, attempted)
+        indices = np.where(indices_raw)[0].tolist()
+        view = CDSView(source=self.source, filters=[IndexFilter(indices)])
+        return view
 
-    @wrappy.todo("Review whether it's appropriate to create a ColumnDataSource")
-    def _source_missed(self, L_labeled, targets):
+    def _view_missed(self, L_labeled, targets):
         """
         Determine the subset missed by a labeling function.
         """
         targetable = np.isin(self.df_labeled["label"], targets)
         abstained = L_labeled == module_config.ABSTAIN_DECODED
-        indices = np.multiply(targetable, abstained)
-        return ColumnDataSource(self.df_labeled.iloc[indices])
+        indices_raw = np.multiply(targetable, abstained)
+        indices = np.where(indices_raw)[0].tolist()
+        view = CDSView(source=self.source, filters=[IndexFilter(indices)])
+        return view
 
-    @wrappy.todo("Review whether it's appropriate to create a ColumnDataSource")
-    def _source_hit(self, L_raw):
+    def _view_hit(self, L_raw):
         """
         Determine the subset hit by a labeling function.
         """
-        indices = L_raw != module_config.ABSTAIN_DECODED
-        return ColumnDataSource(self.df_raw.iloc[indices])
+        indices_raw = L_raw != module_config.ABSTAIN_DECODED
+        indices = np.where(indices_raw)[0].tolist()
+        view = CDSView(source=self.source, filters=[IndexFilter(indices)])
+        return view
